@@ -76,14 +76,14 @@ import Distribution.Verbosity ( normal, Verbosity )
 import Distribution.System ( buildPlatform, Platform )
 
 import Control.Exception ( bracket )
-import Control.Monad ( when, liftM, unless )
+import Control.Monad ( when, liftM, unless, filterM )
 import Data.Char ( toUpper )
 import Data.Monoid ( mempty )
 import System.Directory
     ( createDirectoryIfMissing, doesFileExist, getCurrentDirectory
-    , removeFile )
+    , removeFile, getDirectoryContents )
 import System.Environment ( getEnvironment )
-import System.Exit ( ExitCode(..), exitFailure, exitSuccess, exitWith )
+import System.Exit ( ExitCode(..), exitFailure, exitWith )
 import System.FilePath ( (</>), (<.>) )
 import System.IO ( hClose, IOMode(..), openFile )
 import System.Process ( runProcess, waitForProcess )
@@ -294,7 +294,7 @@ test pkg_descr lbi flags = do
 
     when (not $ PD.hasTests pkg_descr) $ do
         notice verbosity "Package has no test suites."
-        exitSuccess
+        exitWith ExitSuccess
 
     when (PD.hasTests pkg_descr && null enabledTests) $
         die $ "No test suites enabled. Did you remember to configure with "
@@ -315,6 +315,11 @@ test pkg_descr lbi flags = do
 
     createDirectoryIfMissing True testLogDir
 
+    -- Delete ordinary files from test log directory.
+    getDirectoryContents testLogDir
+        >>= filterM doesFileExist . map (testLogDir </>)
+        >>= mapM_ removeFile
+
     let totalSuites = length testsToRun
     notice verbosity $ "Running " ++ show totalSuites ++ " test suites..."
     suites <- mapM doTest testsToRun
