@@ -10,17 +10,16 @@
 #ifndef THREADLABELS_H
 #define THREADLABELS_H
 
-BEGIN_RTS_PRIVATE
+#include "BeginPrivate.h"
 
 #if defined(DEBUG)
 void    initThreadLabelTable (void);
 void    freeThreadLabelTable (void);
-void    updateThreadLabel    (StgWord key, void *data);
 void *  lookupThreadLabel    (StgWord key);
 void    removeThreadLabel    (StgWord key);
 void    labelThread          (StgPtr tso, char *label);
 #endif
 
-END_RTS_PRIVATE
+#include "EndPrivate.h"
 
 #endif /* THREADLABELS_H */
