@@ -11,28 +11,28 @@
 -- Stability   :  experimental
 -- Portability :  portable
 -----------------------------------------------------------------------------
-
 module Haddock.Interface.AttachInstances (attachInstances) where
 
 
 import Haddock.Types
 import Haddock.Convert
 
+import Control.Arrow
 import Data.List
+import qualified Data.Map as Map
 
 import GHC
 import Name
 import InstEnv
 import Class
-import HscTypes (withSession, ioMsg)
-import TcRnDriver (tcRnGetInfo)
-
-#if __GLASGOW_HASKELL__ > 610 || (__GLASGOW_HASKELL__ == 610 && __GHC_PATCHLEVEL__ >= 2)
-import TypeRep hiding (funTyConName)
+#if MIN_VERSION_ghc(7,1,0)
+import GhcMonad (withSession)
 #else
-import TypeRep
+import HscTypes (withSession)
 #endif
-
+import MonadUtils (liftIO)
+import TcRnDriver (tcRnGetInfo)
+import TypeRep hiding (funTyConName)
 import Var hiding (varName)
 import TyCon
 import PrelNames
@@ -40,28 +40,68 @@ import FastString
 #define FSLIT(x) (mkFastString# (x#))
 
 
-attachInstances :: [Interface] -> Ghc [Interface]
-attachInstances = mapM attach
+attachInstances :: [Interface] -> InstIfaceMap -> Ghc [Interface]
+attachInstances ifaces instIfaceMap = mapM attach ifaces
   where
+    -- TODO: take an IfaceMap as input
+    ifaceMap = Map.fromList [ (ifaceMod i, i) | i <- ifaces ]
+
     attach iface = do
-      newItems <- mapM attachExport $ ifaceExportItems iface
+      newItems <- mapM (attachToExportItem iface ifaceMap instIfaceMap)
+                       (ifaceExportItems iface)
       return $ iface { ifaceExportItems = newItems }
 
-    attachExport export@ExportDecl{expItemDecl = L _ (TyClD d)} = do
-       mb_info <- getAllInfo (unLoc (tcdLName d))
-       return $ export { expItemInstances = case mb_info of
-         Just (_, _, instances) ->
-           map synifyInstHead . sortImage instHead . map instanceHead $ instances
-         Nothing ->
-           []
-        }
-    attachExport export = return export
+
+attachToExportItem :: Interface -> IfaceMap -> InstIfaceMap -> ExportItem Name -> Ghc (ExportItem Name)
+attachToExportItem iface ifaceMap instIfaceMap export =
+  case export of
+    ExportDecl { expItemDecl = L _ (TyClD d) } -> do
+      mb_info <- getAllInfo (unLoc (tcdLName d))
+      let export' =
+            export {
+              expItemInstances =
+                case mb_info of
+                  Just (_, _, instances) ->
+                    let insts = map (first synifyInstHead) $ sortImage (first instHead)
+                                [ (instanceHead i, getName i) | i <- instances ]
+                    in [ (inst, lookupInstDoc name iface ifaceMap instIfaceMap)
+                       | (inst, name) <- insts ]
+                  Nothing -> []
+            }
+      return export'
+    _ -> return export
+
+
+lookupInstDoc :: Name -> Interface -> IfaceMap -> InstIfaceMap -> Maybe (Doc Name)
+-- TODO: capture this pattern in a function (when we have streamlined the
+-- handling of instances)
+lookupInstDoc name iface ifaceMap instIfaceMap =
+  case Map.lookup name (ifaceInstanceDocMap iface) of
+    Just doc -> Just doc
+    Nothing ->
+      case Map.lookup modName ifaceMap of
+        Just iface2 ->
+          case Map.lookup name (ifaceInstanceDocMap iface2) of
+            Just doc -> Just doc
+            Nothing -> Nothing
+        Nothing ->
+          case Map.lookup modName instIfaceMap of
+            Just instIface ->
+              case Map.lookup name (instDocMap instIface) of
+                Just (doc, _) -> doc
+                Nothing -> Nothing
+            Nothing -> Nothing
+  where
+    modName = nameModule name
 
 
 -- | Like GHC's getInfo but doesn't cut things out depending on the
 -- interative context, which we don't set sufficiently anyway.
 getAllInfo :: GhcMonad m => Name -> m (Maybe (TyThing,Fixity,[Instance]))
-getAllInfo name = withSession $ \hsc_env -> ioMsg $ tcRnGetInfo hsc_env name
+getAllInfo name = withSession $ \hsc_env -> do 
+   (_msgs, r) <- liftIO $ tcRnGetInfo hsc_env name
+   return r
+
 
 --------------------------------------------------------------------------------
 -- Collecting and sorting instances
